/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "client.h"
#include "assert/assert.h"
#include "util/memory.h"
#include "console/console.h"
#include "network/server/server.h"

client_t CLIENT;

void cmdJoin(const consolecmdexec_t *exec) {
  clientconnect_t connect;

  if(exec->argc > 0) {
    connect.type = CLIENT_TYPE_NETWORKED;
    connect.networked.port = SERVER_DEFAULT_PORT;
  } else {
    connect.type = CLIENT_TYPE_SINGLE_PLAYER;
  }

  errorret_t ret = clientConnect(connect);
  if(ret == ERROR_OK) {
    consolePrint("Connected to server");
  } else {
    consolePrint("Failed to connect to server: %s", errorString());
  }
  errorFlush();
}

void cmdLeave(const consolecmdexec_t *exec) {
  clientDisconnect();
}

void clientInit() {
  memoryZero(&CLIENT, sizeof(client_t));
  consoleRegCmd("join", cmdJoin);
  consoleRegCmd("leave", cmdLeave);
}

errorret_t clientConnect(const clientconnect_t connect) {
  errorret_t ret;

  // Don't connect if already connected.
  if(CLIENT.state != CLIENT_STATE_DISCONNECTED) {
    return error("Client is already connected");
  }

  // Init the client.
  CLIENT.type = connect.type;
  packetQueueInit(&CLIENT.packetQueue);

  // Change how we connect based on the type.
  switch(connect.type) {
    case CLIENT_TYPE_NETWORKED:
      ret = networkedClientConnect(&CLIENT, connect.networked);
      break;

    default:
      assertUnreachable("Invalid client type");
  }

  // Should be connected now.
  if(ret != ERROR_OK) CLIENT.state = CLIENT_STATE_DISCONNECTED;
  return ret;
}

void clientUpdate() {
  packet_t packet;
  int32_t ret;
  errorret_t err;

  // Don't do anything if not connected
  if(CLIENT.state == CLIENT_STATE_DISCONNECTED) return;

  do {
    // Pop the packet off the queue.
    ret = packetQueuePopIn(
      &CLIENT.packetQueue,
      &packet
    );

    // No more packets to process, just break out.
    if(ret == 0) break;

    // An error occured?
    if(ret < 0) {
      clientDisconnect();
      consolePrint("Failed to pop packet");
      return;
    }

    // Process the packet
    err = packetClientProcess(&packet, &CLIENT);
    if(err != ERROR_OK) {
      clientDisconnect();
      consolePrint("Failed to process packet %s", errorString());
      errorFlush();
      return;
    }
  } while(true);
}

void clientDisconnect() {
  // Don't disconnect if already disconnected.
  if(CLIENT.state == CLIENT_STATE_DISCONNECTED) return;

  // Disconnect the client.
  switch(CLIENT.type) {
    case CLIENT_TYPE_NETWORKED:
      networkedClientDisconnect(&CLIENT);
      break;

    default:
      assertUnreachable("Invalid client type");
  }
  
  consolePrint("Client disconnected");
}

void clientDispose() {
  clientDisconnect();
}