/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "network/server/server.h"
#include "assert/assert.h"
#include "util/memory.h"
#include "console/console.h"

errorret_t networkedServerStart(
  server_t *server,
  const serverstart_t start
) {
  assertNotNull(server, "Server is null");
  assertTrue(start.type == SERVER_TYPE_NETWORKED, "Invalid server type");
  assertTrue(server->state == SERVER_STATE_STOPPED,"Server is already running");
  assertIsMainThread("Server start must be on main thread");

  // Initialize the networked server.
  consolePrint("Starting server on port %d...", start.networked.port);
  memoryZero(&server->networked, sizeof(server->networked));

  // Initialize the server socket
  server->networked.socket = socket(AF_INET, SOCK_STREAM, 0);
  if(server->networked.socket < 0) {
    return error("Failed to create socket");
  }

  // Setup the socket options
  server->networked.address.sin_addr.s_addr = INADDR_ANY;
  server->networked.address.sin_family = AF_INET;
  server->networked.address.sin_port = htons(start.networked.port);

  // Set the socket to reuse the address
  int opt = 1;
  setsockopt(
    server->networked.socket,
    SOL_SOCKET,
    SO_REUSEADDR,
    &opt,
    sizeof(opt)
  );

  // Set the socket to non-blocking
  int flags = fcntl(server->networked.socket, F_GETFL, 0);
  if(
    flags < 0 ||
    fcntl(server->networked.socket, F_SETFL, flags | O_NONBLOCK) < 0
  ) {
    close(server->networked.socket);
    return error("Failed to set socket to non-blocking %s", strerror(errno));
  }

  // Bind the socket to the address and port
  int32_t res = bind(
    server->networked.socket,
    (struct sockaddr *)&server->networked.address,
    sizeof(server->networked.address)
  );
  if(res != 0) {
    close(server->networked.socket);
    return error(
      "Socket bind to port %d failed: %s",
      start.networked.port,
      strerror(errno)
    );
  }

  // Set the socket to listen for incoming connections
  res = listen(server->networked.socket, SERVER_MAX_CLIENTS);
  if(res != 0) {
    close(server->networked.socket);
    return error(
      "Failed to listen on port %d: %s",
      start.networked.port,
      strerror(errno)
    );
  }


  // Start the server thread.
  pthread_mutex_init(&server->networked.lock, NULL);
  server->state = SERVER_STATE_STARTING;

  res = pthread_create(&server->thread, NULL, networkedServerThread, server);
  if(res != 0) {
    close(server->networked.socket);
    pthread_mutex_destroy(&server->networked.lock);
    server->state = SERVER_STATE_STOPPED;
    return error("Failed to create server thread");
  }

  // Server started, hand back.
  consolePrint("Server started.");
  return ERROR_OK;
}

void networkedServerStop(server_t *server) {
  assertNotNull(server, "Server is null");
  assertTrue(server->state != SERVER_STATE_STOPPED,"Server is already stopped");
  assertTrue(server->state != SERVER_STATE_STARTING, "Server is starting");
  assertIsMainThread("Server stop must be on main thread");

  // Notify thread to stop
  pthread_mutex_lock(&server->networked.lock);
  server->state = SERVER_STATE_STOPPING;
  pthread_mutex_unlock(&server->networked.lock);

  // Disconnect clients
  uint8_t i = 0;
  do {
    serverclient_t *client = &server->clients[i++];
    if(client->state == SERVER_CLIENT_STATE_DISCONNECTED) continue;
    serverClientClose(client);
  } while(i < SERVER_MAX_CLIENTS);

  // Wait for the server thread to finish
  pthread_join(server->thread, NULL);
  pthread_mutex_destroy(&server->networked.lock);

  close(server->networked.socket);

  server->networked.socket = -1;
  server->state = SERVER_STATE_STOPPED;
  consolePrint("Server stopped.");
}

void * networkedServerThread(void *arg) {
  assertNotNull(arg, "Server is null");
  assertNotMainThread("Server thread must not be main thread");

  server_t *server = (server_t *)arg;

  // Set server state to running
  pthread_mutex_lock(&server->networked.lock);
  server->state = SERVER_STATE_RUNNING;
  pthread_mutex_unlock(&server->networked.lock);

  fd_set readfds;
  struct timeval timeout;

  // Main server loop
  while (1) {
    // Check if the server should continue running
    pthread_mutex_lock(&server->networked.lock);
    if (server->state != SERVER_STATE_RUNNING) {
      pthread_mutex_unlock(&server->networked.lock);
      break;
    }
    pthread_mutex_unlock(&server->networked.lock);

    // Prepare the select call to wait for new connections
    FD_ZERO(&readfds);
    FD_SET(server->networked.socket, &readfds);
    timeout.tv_sec = 1;
    timeout.tv_usec = 0;

    // Wait for incoming connections
    int activity = select(
      server->networked.socket + 1,
      &readfds,
      NULL,
      NULL,
      &timeout
    );
    if (activity <= 0) continue;
    if (!FD_ISSET(server->networked.socket, &readfds)) continue;

    // Accept incoming connection
    int32_t clientSocket = accept(server->networked.socket, NULL, NULL);
    if (clientSocket < 0) {
      consolePrint("Failed to accept connection");
      continue;
    }

    // Find an available client slot
    serverclient_t *client = NULL;
    for (uint8_t i = 0; i < SERVER_MAX_CLIENTS; i++) {
      if(server->clients[i].state != SERVER_CLIENT_STATE_DISCONNECTED) {
        continue;
      }
      client = &server->clients[i];
      break;
    }

    // Server full, send disconnect packet
    if (!client) {
      packet_t packet;
      serverclient_t tempClient = {
        .networked = { .socket = clientSocket },
        .server = server,
        .state = SERVER_CLIENT_STATE_ACCEPTING
      };
      packetDisconnectCreate(&packet, PACKET_DISCONNECT_REASON_SERVER_FULL);
      networkedServerClientWritePacket(&tempClient, &packet);
      if (errorCheck()) errorPrint();
      consolePrint("Client disconnected: Server full.");
      close(clientSocket);
      continue;
    }

    // Accept the client into the server
    errorret_t ret = serverClientAccept(client, (serverclientaccept_t){
      .server = server,
      .networked = { .socket = clientSocket }
    });
    if (ret != ERROR_OK) {
      consolePrint("Failed to accept client connection: %s", errorString());
      errorFlush();
      close(clientSocket);
    }
  }

  // Cleanup and signal server stop
  pthread_mutex_lock(&server->networked.lock);
  server->state = SERVER_STATE_STOPPED;
  pthread_mutex_unlock(&server->networked.lock);

  consolePrint("Server thread exiting.");
  return NULL;
}
